"""
cifar.py
========
Contain nets for the cifar dataset
"""

import logging

import torch.nn as nn
import torch.nn.functional as F
import numpy as np

import indexedconv.utils as utils
import indexedconv.engine as engine


class WideNetIndexConvIndexPool(nn.Module):
    """ResNet like Network from HexaConv paper implemented with indexed convolutions and pooling.

    Args:
        index_matrix (`torch.Tensor`): The index matrix corresponding to the input images.
        camera_layout (str): The grid shape of the images.
    """

    def __init__(self, index_matrix, camera_layout):
        super(WideNetIndexConvIndexPool, self).__init__()

        self.camera_layout = camera_layout

        index_matrix0 = index_matrix

        if self.camera_layout == 'Hex':
            n1 = 17
            n2 = 35
            n3 = 69
        elif self.camera_layout == 'Square':
            n1 = 15
            n2 = 31
            n3 = 61

        # Layer 0
        indices_conv0 = utils.neighbours_extraction(index_matrix0,
                                                    kernel_type=self.camera_layout)
        self.cv0 = engine.IndexedConv(3, n1, indices_conv0)

        # Layer 1 : IndexedConv
        index_matrix1 = index_matrix0
        indices_conv1 = utils.neighbours_extraction(index_matrix1,
                                                    kernel_type=self.camera_layout)
        self.cv1_1 = engine.IndexedConv(n1, n1, indices_conv1)
        self.cv1_2 = engine.IndexedConv(n1, n1, indices_conv1)
        self.cv1_3 = engine.IndexedConv(n1, n1, indices_conv1)
        self.cv1_4 = engine.IndexedConv(n1, n1, indices_conv1)
        self.cv1_5 = engine.IndexedConv(n1, n1, indices_conv1)
        self.cv1_6 = engine.IndexedConv(n1, n1, indices_conv1)
        self.cv1_7 = engine.IndexedConv(n1, n1, indices_conv1)
        self.cv1_8 = engine.IndexedConv(n1, n1, indices_conv1)

        self.bn1_1 = nn.BatchNorm1d(n1)
        self.bn1_2 = nn.BatchNorm1d(n1)
        self.bn1_3 = nn.BatchNorm1d(n1)
        self.bn1_4 = nn.BatchNorm1d(n1)
        self.bn1_5 = nn.BatchNorm1d(n1)
        self.bn1_6 = nn.BatchNorm1d(n1)
        self.bn1_7 = nn.BatchNorm1d(n1)
        self.bn1_8 = nn.BatchNorm1d(n1)

        indices_res_conv1 = utils.neighbours_extraction(index_matrix1,
                                                        kernel_type='Square', radius=0, stride=2)
        self.res_cv1to2 = engine.IndexedConv(n1, n2, indices_res_conv1)

        # Layer 2 : IndexedConv
        indices_conv1to2 = utils.neighbours_extraction(index_matrix1,
                                                       kernel_type=self.camera_layout, stride=2)
        self.cv2_1 = engine.IndexedConv(n1, n2, indices_conv1to2)

        index_matrix2 = utils.pool_index_matrix(index_matrix1, kernel_type=self.camera_layout, stride=2)
        indices_conv2 = utils.neighbours_extraction(index_matrix2,
                                                    kernel_type=self.camera_layout)
        self.cv2_2 = engine.IndexedConv(n2, n2, indices_conv2)
        self.cv2_3 = engine.IndexedConv(n2, n2, indices_conv2)
        self.cv2_4 = engine.IndexedConv(n2, n2, indices_conv2)
        self.cv2_5 = engine.IndexedConv(n2, n2, indices_conv2)
        self.cv2_6 = engine.IndexedConv(n2, n2, indices_conv2)
        self.cv2_7 = engine.IndexedConv(n2, n2, indices_conv2)
        self.cv2_8 = engine.IndexedConv(n2, n2, indices_conv2)

        self.bn2_1 = nn.BatchNorm1d(n1)
        self.bn2_2 = nn.BatchNorm1d(n2)
        self.bn2_3 = nn.BatchNorm1d(n2)
        self.bn2_4 = nn.BatchNorm1d(n2)
        self.bn2_5 = nn.BatchNorm1d(n2)
        self.bn2_6 = nn.BatchNorm1d(n2)
        self.bn2_7 = nn.BatchNorm1d(n2)
        self.bn2_8 = nn.BatchNorm1d(n2)

        indices_res_conv2 = utils.neighbours_extraction(index_matrix2,
                                                        kernel_type='Square', radius=0, stride=2)
        self.res_cv2to3 = engine.IndexedConv(n2, n3, indices_res_conv2)

        # Layer 3 : IndexedConv
        indices_conv2to3 = utils.neighbours_extraction(index_matrix2,
                                                       kernel_type=self.camera_layout, stride=2)
        self.cv3_1 = engine.IndexedConv(n2, n3, indices_conv2to3)
        index_matrix3 = utils.pool_index_matrix(index_matrix2, kernel_type=self.camera_layout, stride=2)
        indices_conv3 = utils.neighbours_extraction(index_matrix3,
                                                    kernel_type=self.camera_layout)
        self.cv3_2 = engine.IndexedConv(n3, n3, indices_conv3)
        self.cv3_3 = engine.IndexedConv(n3, n3, indices_conv3)
        self.cv3_4 = engine.IndexedConv(n3, n3, indices_conv3)
        self.cv3_5 = engine.IndexedConv(n3, n3, indices_conv3)
        self.cv3_6 = engine.IndexedConv(n3, n3, indices_conv3)
        self.cv3_7 = engine.IndexedConv(n3, n3, indices_conv3)
        self.cv3_8 = engine.IndexedConv(n3, n3, indices_conv3)

        self.bn3_1 = nn.BatchNorm1d(n2)
        self.bn3_2 = nn.BatchNorm1d(n3)
        self.bn3_3 = nn.BatchNorm1d(n3)
        self.bn3_4 = nn.BatchNorm1d(n3)
        self.bn3_5 = nn.BatchNorm1d(n3)
        self.bn3_6 = nn.BatchNorm1d(n3)
        self.bn3_7 = nn.BatchNorm1d(n3)
        self.bn3_8 = nn.BatchNorm1d(n3)

        self.bn4 = nn.BatchNorm1d(n3)

        self.conv4 = nn.Conv1d(in_channels=n3, out_channels=10, kernel_size=1, stride=1)

        for m in self.modules():
            if isinstance(m, engine.IndexedConv):
                m.weight.data.normal_(0.0, np.sqrt(2 / (m.kernel_size * m.in_channels)))

    def forward(self, x):
        dropout = nn.Dropout(p=0)

        out = self.cv0(x)
        res1 = out
        out = dropout(self.cv1_1(F.relu(self.bn1_1(out))))
        out = self.cv1_2(F.relu(self.bn1_2(out)))
        res2 = out + res1
        out = dropout(self.cv1_3(F.relu(self.bn1_3(res2))))
        out = self.cv1_4(F.relu(self.bn1_4(out)))
        res3 = out + res2
        out = dropout(self.cv1_5(F.relu(self.bn1_5(res3))))
        out = self.cv1_6(F.relu(self.bn1_6(out)))
        res4 = out + res3
        out = dropout(self.cv1_7(F.relu(self.bn1_7(res4))))
        out = self.cv1_8(F.relu(self.bn1_8(out)))
        res4_1 = res4 + out
        res4_2 = self.res_cv1to2(res4_1)  # Residual link down sampled

        out = dropout(self.cv2_1(F.relu(self.bn2_1(res4_1))))
        out = self.cv2_2(F.relu(self.bn2_2(out)))
        res5 = out + res4_2
        out = dropout(self.cv2_3(F.relu(self.bn2_3(res5))))
        out = self.cv2_4(F.relu(self.bn2_4(out)))
        res6 = out + res5
        out = dropout(self.cv2_5(F.relu(self.bn2_5(res6))))
        out = self.cv2_6(F.relu(self.bn2_6(out)))
        res7 = out + res6
        out = dropout(self.cv2_7(F.relu(self.bn2_7(res7))))
        out = self.cv2_8(F.relu(self.bn2_8(out)))
        res8 = out + res7
        res8_2 = self.res_cv2to3(res8)  # Residual link down sampled

        out = dropout(self.cv3_1(F.relu(self.bn3_1(res8))))
        out = self.cv3_2(F.relu(self.bn3_2(out)))
        res9 = out + res8_2
        out = dropout(self.cv3_3(F.relu(self.bn3_3(res9))))
        out = self.cv3_4(F.relu(self.bn3_4(out)))
        res10 = out + res9
        out = dropout(self.cv3_5(F.relu(self.bn3_5(res10))))
        out = self.cv3_6(F.relu(self.bn3_6(out)))
        res11 = out + res10
        out = dropout(self.cv3_7(F.relu(self.bn3_7(res11))))
        out = self.cv3_8(F.relu(self.bn3_8(out)))
        res12 = out + res11

        out = F.relu(self.bn4(res12))
        out = F.avg_pool1d(out, kernel_size=out.shape[2], stride=1, padding=0)

        out = self.conv4(out)

        out = out.view(out.size(0), -1)

        return F.log_softmax(out, dim=1)


class WideNetIndexConvIndexPoolRetina(nn.Module):
    """ResNet like Network from HexaConv paper implemented with indexed convolutions (retina like kernel) and pooling.

    Args:
        index_matrix (`torch.Tensor`): The index matrix corresponding to the input images.
        camera_layout (str): The grid shape of the images.
    """

    def __init__(self, index_matrix, camera_layout):
        super(WideNetIndexConvIndexPoolRetina, self).__init__()

        self.camera_layout = camera_layout
        index_matrix0 = index_matrix

        if self.camera_layout == 'Hex':
            n1 = 17
            n2 = 35
            n3 = 69
        elif self.camera_layout == 'Square':
            n1 = 15
            n2 = 31
            n3 = 61

        # Layer 0
        indices_conv0 = utils.neighbours_extraction(index_matrix0,
                                                    kernel_type=self.camera_layout, radius=3, retina=True)
        self.cv0 = engine.IndexedConv(3, n1, indices_conv0)

        # Layer 1 : IndexedConv
        index_matrix1 = index_matrix0
        indices_conv1 = utils.neighbours_extraction(index_matrix1,
                                                    kernel_type=self.camera_layout, radius=3, retina=True)
        self.cv1_1 = engine.IndexedConv(n1, n1, indices_conv1)
        self.cv1_2 = engine.IndexedConv(n1, n1, indices_conv1)
        self.cv1_3 = engine.IndexedConv(n1, n1, indices_conv1)
        self.cv1_4 = engine.IndexedConv(n1, n1, indices_conv1)
        self.cv1_5 = engine.IndexedConv(n1, n1, indices_conv1)
        self.cv1_6 = engine.IndexedConv(n1, n1, indices_conv1)
        self.cv1_7 = engine.IndexedConv(n1, n1, indices_conv1)
        self.cv1_8 = engine.IndexedConv(n1, n1, indices_conv1)

        self.bn1_1 = nn.BatchNorm1d(n1)
        self.bn1_2 = nn.BatchNorm1d(n1)
        self.bn1_3 = nn.BatchNorm1d(n1)
        self.bn1_4 = nn.BatchNorm1d(n1)
        self.bn1_5 = nn.BatchNorm1d(n1)
        self.bn1_6 = nn.BatchNorm1d(n1)
        self.bn1_7 = nn.BatchNorm1d(n1)
        self.bn1_8 = nn.BatchNorm1d(n1)

        indices_res_conv1 = utils.neighbours_extraction(index_matrix1,
                                                        kernel_type='Square', radius=0, stride=2)
        self.res_cv1to2 = engine.IndexedConv(n1, n2, indices_res_conv1)

        # Layer 2 : IndexedConv
        indices_conv1to2 = utils.neighbours_extraction(index_matrix1,
                                                       kernel_type=self.camera_layout, stride=2, radius=3, retina=True)
        self.cv2_1 = engine.IndexedConv(n1, n2, indices_conv1to2)

        index_matrix2 = utils.pool_index_matrix(index_matrix1, kernel_type=self.camera_layout, stride=2)
        indices_conv2 = utils.neighbours_extraction(index_matrix2,
                                                    kernel_type=self.camera_layout, radius=3, retina=True)
        self.cv2_2 = engine.IndexedConv(n2, n2, indices_conv2)
        self.cv2_3 = engine.IndexedConv(n2, n2, indices_conv2)
        self.cv2_4 = engine.IndexedConv(n2, n2, indices_conv2)
        self.cv2_5 = engine.IndexedConv(n2, n2, indices_conv2)
        self.cv2_6 = engine.IndexedConv(n2, n2, indices_conv2)
        self.cv2_7 = engine.IndexedConv(n2, n2, indices_conv2)
        self.cv2_8 = engine.IndexedConv(n2, n2, indices_conv2)

        self.bn2_1 = nn.BatchNorm1d(n1)
        self.bn2_2 = nn.BatchNorm1d(n2)
        self.bn2_3 = nn.BatchNorm1d(n2)
        self.bn2_4 = nn.BatchNorm1d(n2)
        self.bn2_5 = nn.BatchNorm1d(n2)
        self.bn2_6 = nn.BatchNorm1d(n2)
        self.bn2_7 = nn.BatchNorm1d(n2)
        self.bn2_8 = nn.BatchNorm1d(n2)

        indices_res_conv2 = utils.neighbours_extraction(index_matrix2,
                                                        kernel_type='Square', radius=0, stride=2)
        self.res_cv2to3 = engine.IndexedConv(n2, n3, indices_res_conv2)

        # Layer 3 : IndexedConv
        indices_conv2to3 = utils.neighbours_extraction(index_matrix2,
                                                       kernel_type=self.camera_layout, stride=2, radius=3, retina=True)
        self.cv3_1 = engine.IndexedConv(n2, n3, indices_conv2to3)
        index_matrix3 = utils.pool_index_matrix(index_matrix2, kernel_type=self.camera_layout, stride=2)
        indices_conv3 = utils.neighbours_extraction(index_matrix3,
                                                    kernel_type=self.camera_layout, radius=3, retina=True)
        self.cv3_2 = engine.IndexedConv(n3, n3, indices_conv3)
        self.cv3_3 = engine.IndexedConv(n3, n3, indices_conv3)
        self.cv3_4 = engine.IndexedConv(n3, n3, indices_conv3)
        self.cv3_5 = engine.IndexedConv(n3, n3, indices_conv3)
        self.cv3_6 = engine.IndexedConv(n3, n3, indices_conv3)
        self.cv3_7 = engine.IndexedConv(n3, n3, indices_conv3)
        self.cv3_8 = engine.IndexedConv(n3, n3, indices_conv3)

        self.bn3_1 = nn.BatchNorm1d(n2)
        self.bn3_2 = nn.BatchNorm1d(n3)
        self.bn3_3 = nn.BatchNorm1d(n3)
        self.bn3_4 = nn.BatchNorm1d(n3)
        self.bn3_5 = nn.BatchNorm1d(n3)
        self.bn3_6 = nn.BatchNorm1d(n3)
        self.bn3_7 = nn.BatchNorm1d(n3)
        self.bn3_8 = nn.BatchNorm1d(n3)

        self.bn4 = nn.BatchNorm1d(n3)

        self.conv4 = nn.Conv1d(in_channels=n3, out_channels=10, kernel_size=1, stride=1)

        for m in self.modules():
            if isinstance(m, engine.IndexedConv):
                m.weight.data.normal_(0.0, np.sqrt(2 / (m.kernel_size * m.in_channels)))

    def forward(self, x):
        dropout = nn.Dropout(p=0)

        out = self.cv0(x)
        res1 = out
        out = dropout(self.cv1_1(F.relu(self.bn1_1(out))))
        out = self.cv1_2(F.relu(self.bn1_2(out)))
        res2 = out + res1
        out = dropout(self.cv1_3(F.relu(self.bn1_3(res2))))
        out = self.cv1_4(F.relu(self.bn1_4(out)))
        res3 = out + res2
        out = dropout(self.cv1_5(F.relu(self.bn1_5(res3))))
        out = self.cv1_6(F.relu(self.bn1_6(out)))
        res4 = out + res3
        out = dropout(self.cv1_7(F.relu(self.bn1_7(res4))))
        out = self.cv1_8(F.relu(self.bn1_8(out)))
        res4_1 = res4 + out
        res4_2 = self.res_cv1to2(res4_1)  # Residual link down sampled

        out = dropout(self.cv2_1(F.relu(self.bn2_1(res4_1))))
        out = self.cv2_2(F.relu(self.bn2_2(out)))
        res5 = out + res4_2
        out = dropout(self.cv2_3(F.relu(self.bn2_3(res5))))
        out = self.cv2_4(F.relu(self.bn2_4(out)))
        res6 = out + res5
        out = dropout(self.cv2_5(F.relu(self.bn2_5(res6))))
        out = self.cv2_6(F.relu(self.bn2_6(out)))
        res7 = out + res6
        out = dropout(self.cv2_7(F.relu(self.bn2_7(res7))))
        out = self.cv2_8(F.relu(self.bn2_8(out)))
        res8 = out + res7
        res8_2 = self.res_cv2to3(res8)  # Residual link down sampled

        out = dropout(self.cv3_1(F.relu(self.bn3_1(res8))))
        out = self.cv3_2(F.relu(self.bn3_2(out)))
        res9 = out + res8_2
        out = dropout(self.cv3_3(F.relu(self.bn3_3(res9))))
        out = self.cv3_4(F.relu(self.bn3_4(out)))
        res10 = out + res9
        out = dropout(self.cv3_5(F.relu(self.bn3_5(res10))))
        out = self.cv3_6(F.relu(self.bn3_6(out)))
        res11 = out + res10
        out = dropout(self.cv3_7(F.relu(self.bn3_7(res11))))
        out = self.cv3_8(F.relu(self.bn3_8(out)))
        res12 = out + res11

        out = F.relu(self.bn4(res12))
        out = F.avg_pool1d(out, kernel_size=out.shape[2], stride=1, padding=0)

        out = self.conv4(out)

        out = out.view(out.size(0), -1)

        return F.log_softmax(out, dim=1)


class WideNet(nn.Module):
    """ResNet like Network from HexaConv paper (Z²).

    """

    def __init__(self):
        super(WideNet, self).__init__()

        n1 = 15
        n2 = 31
        n3 = 61

        # Layer 0
        self.cv0 = nn.Conv2d(3, n1, kernel_size=3, stride=1, padding=1)

        # Layer 1
        self.cv1_1 = nn.Conv2d(n1, n1, kernel_size=3, stride=1, padding=1)
        self.cv1_2 = nn.Conv2d(n1, n1, kernel_size=3, stride=1, padding=1)
        self.cv1_3 = nn.Conv2d(n1, n1, kernel_size=3, stride=1, padding=1)
        self.cv1_4 = nn.Conv2d(n1, n1, kernel_size=3, stride=1, padding=1)
        self.cv1_5 = nn.Conv2d(n1, n1, kernel_size=3, stride=1, padding=1)
        self.cv1_6 = nn.Conv2d(n1, n1, kernel_size=3, stride=1, padding=1)
        self.cv1_7 = nn.Conv2d(n1, n1, kernel_size=3, stride=1, padding=1)
        self.cv1_8 = nn.Conv2d(n1, n1, kernel_size=3, stride=1, padding=1)

        self.bn1_1 = nn.BatchNorm2d(n1)
        self.bn1_2 = nn.BatchNorm2d(n1)
        self.bn1_3 = nn.BatchNorm2d(n1)
        self.bn1_4 = nn.BatchNorm2d(n1)
        self.bn1_5 = nn.BatchNorm2d(n1)
        self.bn1_6 = nn.BatchNorm2d(n1)
        self.bn1_7 = nn.BatchNorm2d(n1)
        self.bn1_8 = nn.BatchNorm2d(n1)

        self.res_cv1to2 = nn.Conv2d(n1, n2, kernel_size=1, stride=2)

        # Layer 2
        self.cv2_1 = nn.Conv2d(n1, n2, kernel_size=3, stride=2, padding=1)

        self.cv2_2 = nn.Conv2d(n2, n2, kernel_size=3, stride=1, padding=1)
        self.cv2_3 = nn.Conv2d(n2, n2, kernel_size=3, stride=1, padding=1)
        self.cv2_4 = nn.Conv2d(n2, n2, kernel_size=3, stride=1, padding=1)
        self.cv2_5 = nn.Conv2d(n2, n2, kernel_size=3, stride=1, padding=1)
        self.cv2_6 = nn.Conv2d(n2, n2, kernel_size=3, stride=1, padding=1)
        self.cv2_7 = nn.Conv2d(n2, n2, kernel_size=3, stride=1, padding=1)
        self.cv2_8 = nn.Conv2d(n2, n2, kernel_size=3, stride=1, padding=1)

        self.bn2_1 = nn.BatchNorm2d(n1)
        self.bn2_2 = nn.BatchNorm2d(n2)
        self.bn2_3 = nn.BatchNorm2d(n2)
        self.bn2_4 = nn.BatchNorm2d(n2)
        self.bn2_5 = nn.BatchNorm2d(n2)
        self.bn2_6 = nn.BatchNorm2d(n2)
        self.bn2_7 = nn.BatchNorm2d(n2)
        self.bn2_8 = nn.BatchNorm2d(n2)

        self.res_cv2to3 = nn.Conv2d(n2, n3, kernel_size=1, stride=2)

        # Layer 3
        self.cv3_1 = nn.Conv2d(n2, n3, kernel_size=3, stride=2, padding=1)

        self.cv3_2 = nn.Conv2d(n3, n3, kernel_size=3, stride=1, padding=1)
        self.cv3_3 = nn.Conv2d(n3, n3, kernel_size=3, stride=1, padding=1)
        self.cv3_4 = nn.Conv2d(n3, n3, kernel_size=3, stride=1, padding=1)
        self.cv3_5 = nn.Conv2d(n3, n3, kernel_size=3, stride=1, padding=1)
        self.cv3_6 = nn.Conv2d(n3, n3, kernel_size=3, stride=1, padding=1)
        self.cv3_7 = nn.Conv2d(n3, n3, kernel_size=3, stride=1, padding=1)
        self.cv3_8 = nn.Conv2d(n3, n3, kernel_size=3, stride=1, padding=1)

        self.bn3_1 = nn.BatchNorm2d(n2)
        self.bn3_2 = nn.BatchNorm2d(n3)
        self.bn3_3 = nn.BatchNorm2d(n3)
        self.bn3_4 = nn.BatchNorm2d(n3)
        self.bn3_5 = nn.BatchNorm2d(n3)
        self.bn3_6 = nn.BatchNorm2d(n3)
        self.bn3_7 = nn.BatchNorm2d(n3)
        self.bn3_8 = nn.BatchNorm2d(n3)

        self.bn4 = nn.BatchNorm2d(n3)

        self.conv4 = nn.Conv1d(in_channels=n3, out_channels=10, kernel_size=1, stride=1)

        for m in self.modules():
            if isinstance(m, nn.Conv2d) | isinstance(m, nn.Conv1d):
                n = m.in_channels
                for k in m.kernel_size:
                    n *= k
                m.weight.data.normal_(0.0, np.sqrt(2 / n))

    def forward(self, x):
        dropout = nn.Dropout(p=0)

        out = self.cv0(x)
        res1 = out
        out = dropout(self.cv1_1(F.relu(self.bn1_1(out))))
        out = self.cv1_2(F.relu(self.bn1_2(out)))
        res2 = out + res1
        out = dropout(self.cv1_3(F.relu(self.bn1_3(res2))))
        out = self.cv1_4(F.relu(self.bn1_4(out)))
        res3 = out + res2
        out = dropout(self.cv1_5(F.relu(self.bn1_5(res3))))
        out = self.cv1_6(F.relu(self.bn1_6(out)))
        res4 = out + res3
        out = dropout(self.cv1_7(F.relu(self.bn1_7(res4))))
        out = self.cv1_8(F.relu(self.bn1_8(out)))
        res4_1 = res4 + out
        res4_2 = self.res_cv1to2(res4_1)  # Residual link down sampled

        out = dropout(self.cv2_1(F.relu(self.bn2_1(res4_1))))
        out = self.cv2_2(F.relu(self.bn2_2(out)))
        res5 = out + res4_2
        out = dropout(self.cv2_3(F.relu(self.bn2_3(res5))))
        out = self.cv2_4(F.relu(self.bn2_4(out)))
        res6 = out + res5
        out = dropout(self.cv2_5(F.relu(self.bn2_5(res6))))
        out = self.cv2_6(F.relu(self.bn2_6(out)))
        res7 = out + res6
        out = dropout(self.cv2_7(F.relu(self.bn2_7(res7))))
        out = self.cv2_8(F.relu(self.bn2_8(out)))
        res8 = out + res7
        res8_2 = self.res_cv2to3(res8)  # Residual link down sampled

        out = dropout(self.cv3_1(F.relu(self.bn3_1(res8))))
        out = self.cv3_2(F.relu(self.bn3_2(out)))
        res9 = out + res8_2
        out = dropout(self.cv3_3(F.relu(self.bn3_3(res9))))
        out = self.cv3_4(F.relu(self.bn3_4(out)))
        res10 = out + res9
        out = dropout(self.cv3_5(F.relu(self.bn3_5(res10))))
        out = self.cv3_6(F.relu(self.bn3_6(out)))
        res11 = out + res10
        out = dropout(self.cv3_7(F.relu(self.bn3_7(res11))))
        out = self.cv3_8(F.relu(self.bn3_8(out)))
        res12 = out + res11

        out = F.relu(self.bn4(res12))

        # out = torch.mean(out, dim=2, keepdim=True)
        out = F.avg_pool2d(out, kernel_size=out.shape[2:], stride=1, padding=0)

        out = self.conv4(out.view(out.shape[0], out.shape[1], -1))

        out = out.view(out.size(0), -1)

        return F.log_softmax(out, dim=1)



"""
utils.py
========
Contain utility functions for the indexed convolution
"""

import logging
import subprocess

import numpy as np
import torch
import torch.nn.functional as F
from torch.utils.data import Dataset
import h5py


def get_gpu_usage_map(device_id):
    """Get the current gpu usage.
    Inspired from `gpu usage from
    pytorch <https://discuss.pytorch.org/t/access-gpu-memory-usage-in-pytorch/3192/4>`_

    Args:
        device_id (int): the GPU id as GPU/Unit's 0-based index in the natural enumeration returned  by the driver

    Returns
        dict - usage
            Keys are device ids as integers.
            Values are memory usage as integers in MB, total memory usage as integers in MB, gpu utilization in %.
    """
    result = subprocess.check_output(
        [
            'nvidia-smi', '--id=' + str(device_id), '--query-gpu=memory.used,memory.total,utilization.gpu',
            '--format=csv,nounits,noheader'
        ], encoding='utf-8')
    # Convert lines into a dictionary
    # gpu_usage = [[int(y) for y in x.split(',')] for x in result.strip().split('\n')]
    gpu_usage = [int(y) for y in result.split(',')]
    # gpu_usage_map = {}
    gpu_usage_map = {'memory_used': gpu_usage[0], 'total_memory': gpu_usage[1], 'utilization': gpu_usage[2]}
    # for i, gpu in enumerate(gpu_usage):
    #     gpu_usage_map[i] = {'memory_used': gpu[0], 'total_memory': gpu[1], 'utilization': gpu[2]}
    return gpu_usage_map


def compute_total_parameter_number(net):
    r"""Computes the total number of parameters of a network.

    Args:
        net (:class:`nn.Module`): The network.

    Returns:
        A int
    """
    num_parameters = 0
    for name, param in net.named_parameters():
        num_parameters += param.clone().cpu().data.view(-1).size(0)

    return num_parameters


#####################
# Indexed functions #
#####################

def create_index_matrix(nbRow, nbCol, injTable):
    r"""Creates the matrix of index of the pixels of the images of any shape stored as vectors.

    Args:
        nbRow (int): The number of rows of the index matrix.
        nbCol (int): The number of cols of the index matrix.
        injTable (numpy.array): The injunction table, i.e. the list of the position of every pixels of the vector image
            in a vectorized square image.

    Returns:
        A torch.Tensor containing the index of each pixel represented in a matrix.

    Example:
        >>> image = [0, 1, 2, 3, 4, 5, 6]  # hexagonal image stored as a vector
        >>> # in the hexagonal space                  0 1
        >>> #                                        2 3 4
        >>> #                                         5 6
        >>> # injunction table of the pixel position of a hexagonal image represented in the axial addressing system
        >>> injTable = [0, 1, 3, 4, 5, 7, 8]
        >>> index_matrix = [[0, 1, -1], [2, 3, 4], [-1, 5, 6]]
        [[0, 1, -1],
        [2,  3, 4],
        [-1, 5, 6]]
    """
    logger = logging.getLogger(__name__ + '.create_index_matrix')
    index_matrix = torch.full((int(nbRow), int(nbCol)), -1)
    for i, idx in enumerate(injTable):
        idx_row = int(idx // nbRow)
        idx_col = int(idx % nbCol)
        index_matrix[idx_row,idx_col] = i

    index_matrix.unsqueeze_(0)
    index_matrix.unsqueeze_(0)
    return index_matrix


def img2mat(input_images, index_matrix):
    """Transforms a batch of features of vector images in a batch of features of matrix images.

     Args:
         input_images (torch.Tensor): The images with shape (batch, features, image).
         index_matrix (torch.Tensor): The index matrix containing the index of the pixels of the images.
            represented in a matrix

     Returns:
         A torch.Tensor
     """
    logger = logging.getLogger(__name__ + '.img2mat')
    # First create a tensor of shape : batch, features, index_matrix.size filled with zeros
    image_matrix = input_images.new_zeros((input_images.size(0),
                                           input_images.size(1),
                                           index_matrix.size(-2),
                                           index_matrix.size(-1)), dtype=torch.int)

    logger.debug('image matrix shape : {}'.format(image_matrix.size()))

    for i in range(index_matrix.size(-2)):  # iterate over the rows of index matrix
        for j in range(index_matrix.size(-1)):  # iterate over the cols of index matrix
            if index_matrix.data[0, 0, i, j] != -1:
                image_matrix[:, :, i, j] = input_images[:, :, int(index_matrix[0, 0, i, j])]

    return image_matrix


def mat2img(input_matrix, index_matrix):
    """
    Transforms a batch of features of matrix images in a batch of features of vector images.

    Args:
        input_matrix (torch.Tensor): The images with shape (batch, features, matrix.size).
        index_matrix (torch.Tensor): The index matrix for the images, shape(1, 1, matrix.size).

    """
    logger = logging.getLogger(__name__ + '.mat2img')
    logger.debug('input matrix shape : {}'.format(input_matrix.size()))
    image_length = index_matrix[0, 0, torch.ge(index_matrix[0, 0], 0)].size(0)

    logger.debug('new image length : {}'.format(image_length))

    images = input_matrix.new_zeros((input_matrix.size(0), input_matrix.size(1), image_length), dtype=torch.float)

    logger.debug('new images shape : {}'.format(images.size()))

    for i in range(index_matrix.size(-2)):  # iterate over the rows of index matrix
        for j in range(index_matrix.size(-1)):  # iterate over the cols of index matrix
            if index_matrix[0, 0, i, j] != -1:
                images[:, :, int(index_matrix[0, 0, i, j])] = input_matrix[:, :,  i, j]

    return images


def pool_index_matrix(index_matrix, kernel_type='Pool', stride=2):
    """
    Pools an index matrix.

    Args:
        index_matrix (torch.Tensor): The index matrix for the images, shape(1, 1, matrix.size).
        kernel_type (str): The kernel shape, Hex for hexagonal, Square for a square of size 3
            and Pool for a square of size 2.
        stride (int): The stride.

    Returns:
        A torch.Tensor containing the pooled matrix.
    """
    logger = logging.getLogger(__name__ + '.pool_index_matrix')
    if kernel_type == 'Pool':
        weight = torch.Tensor([[1, 0], [0, 0]]).requires_grad_(False)
        padding = 0
    elif kernel_type == 'Square' or kernel_type == 'Hex':
        weight = torch.Tensor([[0, 0, 0], [0, 1, 0], [0, 0, 0]]).requires_grad_(False)
        padding = 1
    weight.unsqueeze_(0)
    weight.unsqueeze_(0)

    pooled_matrix = F.conv2d(index_matrix, weight, stride=stride, padding=padding).data

    logger.debug('pooled matrix shape : {}'.format(pooled_matrix.size()))
    # Index reconstruction
    idx = 0
    for i in range(pooled_matrix.size(-2)):  # iterate over the rows of index matrix
        for j in range(pooled_matrix.size(-1)):  # iterate over the cols of index matrix
            if pooled_matrix[0, 0, i, j] != -1:
                pooled_matrix[0, 0, i, j] = idx
                idx += 1

    return pooled_matrix


def build_kernel(kernel_type, radius=1, dilation=1):
    """Builds the convolution kernel or mask. (Following the suggestion of Miguel Lallena).

    Args:
        kernel_type (str): The type of kernel. Can be hexagonal ('Hex') or square ('Square').
        radius (int): The radius of the kernel.
        dilation (int): The dilation. A dilation of 1 means no dilation.

    Returns:
         A :class:`np.array`.
    """
    k_size = 2 * radius * dilation + 1
    kernel = np.zeros((k_size, k_size))
    for i in range(0, k_size, dilation):
        for j in range(0, k_size, dilation):
            if kernel_type == 'Square':
                kernel[i, j] = 1
            elif kernel_type == 'Hex':
                kernel[i, j] = int(np.abs(i -j) <= radius * dilation)
            else:
                raise ValueError('Unknown kernel type')

    return kernel


def neighbours_extraction(index_matrix, kernel_type='Hex', radius=1, stride=1, dilation=1, retina=False):
    """Builds the matrix of indices from an index matrix based on a kernel.

    The matrix of indices contains for each pixel of interest its neighbours, including itself.

    Args:
        index_matrix (torch.Tensor): Matrix of index for the images, shape(1, 1, matrix.size).
        kernel_type (str): The kernel shape, Hex for hexagonal Square for a square and Pool for a square of size 2.
        radius (int): The radius of the kernel.
        stride (int): The stride.
        dilation (int): The dilation. A dilation of 1 means no dilation.
        retina (bool): Whether to build a retina like kernel. If True, dilation must be 1.

    Returns:
        A torch.Tensor - the matrix of the neighbours.

    Example:
        >>> index_matrix = [[0, 1, -1], [2, 3, 4], [-1, 5, 6]]
        [[0, 1, -1],
        [2,  3, 4],
        [-1, 5, 6]]
        >>> kernel_type = 'Hex'
        >>> radius = 1
        >>> kernel
        [[1, 1, 0],
        [ 1, 1, 1],
        [ 0, 1, 1]]
        >>> stride = 1
        >>> neighbours = neighbours_extraction(index_matrix, kernel_type, radius, stride)
        [[-1, -1, -1,  0,  1,  2,  3],
        [ -1, -1,  0,  1, -1,  3,  4],
        [ -1,  0, -1,  2,  3, -1,  5],
        [  0,  1,  2,  3,  4,  5,  6],
        [  1, -1,  3,  4, -1,  6, -1],
        [  2,  3, -1,  5,  6, -1, -1],
        [  3,  4,  5,  6, -1, -1, -1]]
    """
    if retina:
        dilation = 1
    padding = radius * dilation * 2
    stride = stride
    bound = radius * dilation * 2 if radius > 0 else 1
    if kernel_type == 'Pool':
        kernel = np.ones((2, 2), dtype=bool)
        stride = 2
        bound = 1
        padding = 0
        center = 0
    elif retina:
        kernel = build_kernel(kernel_type, 1, radius).astype(bool)
        for i in range(1, radius):
            sub_kernel = np.zeros_like(kernel).astype(bool)
            sub_kernel[i:sub_kernel.shape[0]-i, i:sub_kernel.shape[1]-i] = build_kernel(kernel_type, 1, radius - i).astype(bool)
            kernel = kernel + sub_kernel
        center = int((np.count_nonzero(kernel) - 1) / 2)
    else:
        kernel = build_kernel(kernel_type, radius, dilation).astype(bool)
        center = int((np.count_nonzero(kernel) - 1) / 2)

    neighbours = []

    idx_mtx = np.ones((index_matrix.size(-2)+padding, index_matrix.size(-1)+padding), dtype=int) * (-1)
    offset = int(padding/2)
    if offset == 0:
        idx_mtx = index_matrix[0, 0, :, :].numpy()
    else:
        idx_mtx[offset:-offset, offset:-offset] = index_matrix[0, 0, :, :].numpy()

    for i in range(0, idx_mtx.shape[0]-bound, stride):
        for j in range(0, idx_mtx.shape[1]-bound, stride):
            patch = idx_mtx[i:i+kernel.shape[0], j:j+kernel.shape[1]][kernel]
            if patch[center] == -1:
                continue
            neighbours.append(patch)

    neighbours = np.asarray(neighbours).T
    neighbours = torch.from_numpy(neighbours).long()

    return neighbours


def prepare_mask(indices):
    """Prepares the indices and the mask for the GEMM im2col operation.

    Args:
        indices (torch.Tensor): The matrix of indices containing the neighbours of each pixel of interest.

    """
    padded = indices == -1
    new_indices = indices.clone()
    new_indices[padded] = 0

    mask = torch.FloatTensor([1, 0])
    mask = mask[..., padded.long()]

    return new_indices, mask


###################
# Transformations #
###################

def square_to_hexagonal_index_matrix(image):
    """Creates the index matrix of square images in a hexagonal grid (axial addressing system).

    Args:
        image: input tensor of shape (c, n, m)

        """
    index_matrix = torch.ones(image.shape[1],
                              image.shape[2] + int(np.ceil(image.shape[1] / 2))) * -1
    n = 0
    for i in range(image.shape[1]):
        for j in range(image.shape[2]):
            index_matrix[i, j + int(np.ceil(i / 2))] = n
            n += 1
    return index_matrix


def square_to_hexagonal(image):
    """Rough sampling of square images to hexagonal grid

    Args:
        image: image tensor of shape (c, n, m)

    """
    image_tr = image.clone()
    image_tr[:, :, :-1] = image[:, :, 1:]
    image_mean = (image + image_tr) / 2
    image[:, 1::2, :] = image_mean[:, 1::2, :]

    return image.view(image.shape[0], -1)


def build_hexagonal_position(index_matrix):
    """Computes the position of the pixels in the hexagonal grid from the index matrix.

    Args:
        index_matrix (tensor): The index matrix representing the index of each pixel in the axial addressing system.
    """
    pix_positions = []
    for i in range(index_matrix.shape[0]):
        for j in range(index_matrix.shape[1]):
            if not index_matrix[i, j] == -1:
                pix_positions.append([j - i/2, -i])
    return pix_positions


def normalize(images):
    """Normalizes images.

    Args:
        images: image tensor of shape (c, n, m)

    """
    images -= np.mean(images, axis=(1, 2), keepdims=True)
    std = np.sqrt(images.var(axis=(1, 2), ddof=1, keepdims=True))
    std[std < 1e-8] = 1.
    images /= std
    return images


class PCA(object):
    """ Credit E. Hoogeboom http://github.com/ehoogeboom/hexaconv"""
    def __init__(self, D, n_components):
        self.n_components = n_components
        self.U, self.S, self.m = self.fit(D, n_components)

    def fit(self, D, n_components):
        """
        The computation works as follows:

        - The covariance is :code:`C = 1/(n-1) * D * D.T`

        - The eigendecomp of C is: :code:`C = V Sigma V.T`

        - Let :code:`Y = 1/sqrt(n-1) * D`

        - Let :code:`U S V = svd(Y)`,

        - Then the columns of U are the eigenvectors of :code:`Y * Y.T = C`

        - And the singular values S are the sqrts of the eigenvalues of C

        - We can apply PCA by multiplying by U.T
        """

        # We require scaled, zero-mean data to SVD,
        # But we don't want to copy or modify user data
        m = np.mean(D, axis=1)[:, np.newaxis]
        D -= m
        D *= 1.0 / np.sqrt(D.shape[1] - 1)
        U, S, V = np.linalg.svd(D, full_matrices=False)
        D *= np.sqrt(D.shape[1] - 1)
        D += m
        return U[:, :n_components], S[:n_components], m

    def transform(self, D, whiten=False, ZCA=False, regularizer=10 ** (-5)):
        """
        We want to whiten, which can be done by multiplying by :math:`\sigma^{-1/2} U.T`

        Any orthogonal transformation of this is also white,
        and when :code:`ZCA=True` we choose :math:`U \sigma^{-1/2} U.T`
        """
        if whiten:
            # Compute :math:`\sigma^{-1/2} = S^{-1}`,
            # with smoothing for numerical stability
            Sinv = 1.0 / (self.S + regularizer)

            if ZCA:
                # The ZCA whitening matrix
                W = np.dot(self.U,
                           np.dot(np.diag(Sinv),
                                  self.U.T))
            else:
                # The whitening matrix
                W = np.dot(np.diag(Sinv), self.U.T)

        else:
            W = self.U.T

        # Transform
        return np.dot(W, D - self.m)


###################
# Data            #
###################

class NumpyDataset(Dataset):
    """Loads data in a Dataset from :class:`numpy.array`.

    Args:
        data (numpy.array): The data to load.
        labels (numpy.array): The labels of the data.
        transform (callable, optional): A callable or a composition of callable to be applied to the data.
        target_transform (callable, optional): A callable or a composition of callable to be applied to the labels.
    """
    def __init__(self, data, labels, transform=None, target_transform=None):
        self.images = data
        self.labels = labels
        self.transform = transform
        self.target_transform = target_transform

    def __len__(self):
        return len(self.labels)

    def __getitem__(self, item):

        if self.transform:
            self.images[item] = self.transform(self.images[item])
        if self.target_transform:
            self.labels[item] = self.target_transform(self.labels[item])

        return self.images[item], self.labels[item]


class HDF5Dataset(Dataset):
    """Loads data in a Dataset from a HDF5 file.

    Args:
        path (str): The path to the HDF5 file.
        transform (callable, optional): A callable or a composition of callable to be applied to the data.
        target_transform (callable, optional): A callable or a composition of callable to be applied to the labels.
    """
    def __init__(self, path, transform=None, target_transform=None):
        with h5py.File(path, 'r') as f:
            self.images = f['images'][()]
            self.labels = f['labels'][()]
        self.transform = transform
        self.target_transform = target_transform
        self.logger = logging.getLogger(__name__ + 'HDF5Dataset')

    def __len__(self):
        return len(self.labels)

    def __getitem__(self, item):
        image, label = self.images[item], self.labels[item]
        if self.transform:
            image = self.transform(image)
        if self.target_transform:
            label = self.transform(label)

        return image, label


# Transforms
class NumpyToTensor(object):
    """Converts a numpy array to a tensor."""

    def __call__(self, data):

        data = torch.tensor(data)
        return data


class SquareToHexa(object):
    """Converts an image with a square grid to an image with a hexagonal one."""
    def __call__(self, image):
        image = square_to_hexagonal(image)
        # print(sample)
        return image



from .indexed import IndexedConv, IndexedMaxPool2d, IndexedAveragePool2d
from .masked import MaskedConv2d

__all__ = ['IndexedConv',
           'IndexedMaxPool2d',
           'IndexedAveragePool2d',
           'MaskedConv2d']



"""
indexed.py
========
Contain the core functions for the indexed operations
"""

import logging
import math

import torch
import torch.nn as nn
from torch.nn.parameter import Parameter

import indexedconv.utils as utils


class IndexedMaxPool2d(nn.Module):
    r"""
    Compute the Max Pooling 2d operation on a batch of features of vector images wrt a matrix of indices

    Args:
        indices (LongTensor): index tensor of shape (L x kernel_size),
            having on each row the indices of neighbors of each element of the input
            a -1 indicates the absence of a neighbor, which is handled as zero-padding
    Shape:
        - Input: :math:`(N, C, L_{in})`
        - Output: :math:`(N, C, L_{out})`
    """
    def __init__(self, indices):
        super(IndexedMaxPool2d, self).__init__()
        self.logger = logging.getLogger(__name__ + '.IndexedMaxPool2d')
        self.indices = indices
        self.indices, self.mask = utils.prepare_mask(self.indices)
        self.register_buffer('indices_', self.indices)
        self.register_buffer('mask_', self.mask)

    def forward(self, input_images):
        self.logger.debug('Max pool image')

        col = input_images[..., self.indices_] * self.mask_

        out, _ = torch.max(col, 2)

        return out


class IndexedAveragePool2d(nn.Module):
    r"""
    Compute the Average Pooling 2d operation on a batch of features of vector images wrt a matrix of indices

    Args:
        indices (LongTensor): index tensor of shape (L x kernel_size), having on each
                              row the indices of neighbors of each element of the input
                              a -1 indicates the absence of a neighbor, which is handled
                              as zero-padding
    Shape:
        - Input: :math:`(N, C, L_{in})`
        - Output: :math:`(N, C, L_{out})`

    """
    def __init__(self, indices):
        super(IndexedAveragePool2d, self).__init__()
        self.logger = logging.getLogger(__name__ + '.IndexedAveragePool2d')
        self.indices = indices
        self.indices, self.mask = utils.prepare_mask(self.indices)
        self.register_buffer('indices_', self.indices)
        self.register_buffer('mask_', self.mask)

    def forward(self, input_images):
        self.logger.debug('Average pool image')

        col = input_images[..., self.indices_] * self.mask_

        out = torch.mean(col, 2)

        return out


class IndexedConv(nn.Module):
    r"""Applies a convolution over an input tensor where neighborhood relationships
    between elements are explicitly provided via an `indices` tensor.

    The output value of the layer with input size :math:`(N, C_{in}, L)` and output
    :math:`(N, C_{out}, L)` can be described as:

    .. math::

        \begin{array}{ll}
        out(N_i, C_{out_j})  = bias(C_{out_j})
                       + \sum_{{c}=0}^{C_{in}-1}
                         \sum_{{i}=0}^{L-1}
                         \sum_{{k}=0}^{K} weight(C_{out_j}, c, k) * input(N_i, c, indices(i, k))
        \end{array}

    where

    | `indices` is a L x K tensor, where `K` is the size of the convolution kernel,
    | providing the indices of the `K` neighbors of input element `i`.
    | A -1 entry means zero-padding.

    Args:
        in_channels (int): Number of channels in the input tensor

        out_channels (int): Number of channels produced by the convolution

        indices (LongTensor): index tensor of shape (L x kernel_size), having on each
        row the indices of neighbors of each element of the input a -1 indicates the absence of a
        neighbor, which is handled as zero-padding

        bias (bool, optional): If ``True``, adds a learnable bias to the output. Default: ``True``

    Shape:
        - Input: :math:`(N, C_{in}, L)`
        - Output: :math:`(N, C_{out}, L)`

    Attributes:
        weight (Tensor): the learnable weights of the module of shape
            (out_channels, in_channels, kernel_size)
        bias (Tensor):   the learnable bias of the module of shape
            (out_channels)

    Examples::

        >>> indices = (10 * torch.rand(50, 3)).type(torch.LongTensor)
        >>> m = nn.IndexedConv(16, 3, indices)
        >>> input = torch.randn(20, 16, 50)
        >>> output = m(input)
    """

    def __init__(self, in_channels, out_channels, indices, bias=True):
        super(IndexedConv, self).__init__()
        self.logger = logging.getLogger(__name__ + '.IndexedConv')

        groups = 1

        kernel_size = indices.shape[0]

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.indices, self.mask = utils.prepare_mask(indices)
        self.output_width = self.indices.shape[1]

        self.register_buffer('indices_', self.indices)
        self.register_buffer('mask_', self.mask)

        self.weight = Parameter(torch.Tensor(
            out_channels, in_channels // groups, kernel_size))

        if bias:
            self.bias = Parameter(torch.Tensor(out_channels))
        else:
            self.register_parameter('bias', None)

        self.reset_parameters()

    def reset_parameters(self):
        n = self.in_channels * self.kernel_size
        stdv = 1. / math.sqrt(n)
        self.weight.data.uniform_(-stdv, stdv)
        if self.bias is not None:
            self.bias.data.uniform_(-stdv, stdv)

    def forward(self, input):
        nbatch = input.shape[0]
        col = input[..., self.indices_] * self.mask_
        # col is of shape (N, C_in, K, Wo)
        col = col.view(nbatch, -1, self.output_width)
        out = torch.bmm(self.weight.view(self.out_channels, -1).expand(nbatch, -1, -1), col)
        if self.bias is not None:
            out = out + self.bias.unsqueeze(1)

        return out

    def __repr__(self):
        s = '{name} ({in_channels}, {out_channels}, kernel_size={kernel_size}'
        if self.bias is None:
            s += ', bias=False'
        s += ')'
        return s.format(name=self.__class__.__name__, **self.__dict__)



"""
mnist.py
========
Contain the net for the mnist dataset
"""

import torch
import torch.nn as nn
import torch.nn.functional as F

import indexedconv.utils as utils
from indexedconv.engine import IndexedConv, IndexedMaxPool2d


class GLNet2HexaConvForMnist(nn.Module):
    r"""Network with indexed convolutions and pooling (square kernels).
    2 CL (after each conv layer, pooling is executed)
    1 FC

    Args:
        index_matrix (`torch.Tensor`): The index matrix corresponding to the input images.
    """

    def __init__(self, index_matrix):
        super(GLNet2HexaConvForMnist, self).__init__()

        index_matrix1 = index_matrix

        # Layer 1 : IndexedConv
        indices_conv1 = utils.neighbours_extraction(index_matrix1,
                                                    kernel_type='Hex')
        pool_indices1 = utils.neighbours_extraction(index_matrix1,
                                                    kernel_type='Hex', stride=2)
        self.cv1 = IndexedConv(1, 32, indices_conv1)
        self.max_pool1 = IndexedMaxPool2d(pool_indices1)

        # Layer 2 : IndexedConv
        index_matrix2 = utils.pool_index_matrix(index_matrix1, kernel_type='Hex', stride=2)
        indices_conv2 = utils.neighbours_extraction(index_matrix2,
                                                    kernel_type='Hex')
        pool_indices2 = utils.neighbours_extraction(index_matrix2,
                                                    kernel_type='Hex', stride=2)
        self.cv2 = IndexedConv(32, 64, indices_conv2)
        self.max_pool2 = IndexedMaxPool2d(pool_indices2)

        index_matrix3 = utils.pool_index_matrix(index_matrix2, kernel_type='Hex', stride=2)

        # Compute the number of pixels (where idx is not -1 in the index matrix) of the last features
        n_pixels = int(torch.sum(torch.ge(index_matrix3[0, 0], 0)).data)
        self.lin1 = nn.Linear(n_pixels * 64, 1024)

        self.lin2 = nn.Linear(1024, 10)

    def forward(self, x):
        drop = nn.Dropout(p=0.5)
        out = F.relu(self.cv1(x))
        out = self.max_pool1(out)
        out = F.relu(self.cv2(out))
        out = self.max_pool2(out)

        out = out.view(out.size(0), -1)

        out = self.lin1(out)
        out = drop(F.relu(out))

        out = self.lin2(out)

        return F.log_softmax(out, dim=1)



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

exec(open('indexedconv/version.py').read())

setuptools.setup(
    name="IndexedConv",
    version=__version__,
    author="M. Jacquemont, T. Vuillaume, L. Antiga",
    author_email="jacquemont@lapp.in2p3.fr",
    description="An implementation of indexed convolution and pooling",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/IndexedConv/IndexedConv",
    install_requires=[
        "torch>=1.4",
        "torchvision",
        "numpy",
        "matplotlib",
        "h5py",
        "sphinxcontrib-katex"

    ],
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    license='MIT',
)



"""
masked.py
========
Contain the core functions for the masked operations
"""

import torch.nn as nn
import torch
import torch.nn.functional as F

from indexedconv.utils import build_kernel


class MaskedConv2d(nn.Conv2d):

    def forward(self, x):
        r = int((self.kernel_size[0] - 1) / 2)
        mask = torch.from_numpy(build_kernel('Hex', radius=r)).type(torch.float).to(x.device)
        return F.conv2d(x, self.weight * mask, bias=self.bias, stride=self.stride, padding=self.padding)
